from abc import ABC, abstractmethod
from pathlib import Path
import numpy as np

class DataLoader(ABC):
    """Abstract base class for all data loaders (HDF5, ROOT, etc.)."""

    def __init__(self, path: str | Path):
        self.path = Path(path)
        if not self.path.exists():
            raise FileNotFoundError(f"File not found: {self.path}")

    @abstractmethod
    def get_structure(self) -> dict:
        """Return the hierarchical structure of the file.
        Should return nested dicts: {group: {dataset: shape, ...}, ...}
        """

    @abstractmethod
    def get_dataset_shape(self, dataset_path: str) -> tuple[int, ...]:
        """Return the shape of a given dataset without loading its contents."""

    @abstractmethod
    def load_dataset(self, dataset_path: str) -> np.ndarray:
        """Lazily load a dataset as a NumPy array."""

    @abstractmethod
    def get_flat_dataset_paths(self) -> list[str]:
        """Return a flat list of dataset paths suitable for UI dropdowns."""
        pass



from abc import ABC, abstractmethod
from typing import Any


class HEPBrowsePlugin(ABC):
    """
    Base class for all plugins (builtin and external).
    Plugins should subclass this and implement `register`.
    """

    name: str = "Unnamed"
    version: str = "0.0.0"

    def __init__(self) -> None:
        # Optional place for plugin init state
        self.app = None

    def register(self, app: Any) -> None:
        """
        Called once at startup with the App instance.
        Plugins should use `app` to register tabs, settings providers,
        loaders, services, etc.
        """
        self.app = app

    # convenience extension points (optional to implement)
    def register_tabs(self, app: Any) -> None:
        """Register TabPanes via app.add_tab(label, id, factory)."""
        return

    def register_settings(self, settings_manager: Any) -> None:
        """Register settings providers if any."""
        return



# src/hepbrowse/data/h5_loader.py
from pathlib import Path
import h5py
import numpy as np

from hepbrowse.data.loader_base import DataLoader


class H5Loader(DataLoader):
    """Thread-safe HDF5 loader."""

    def get_structure(self) -> dict:
        def explore(obj):
            struct = {}
            for name, item in obj.items():
                if isinstance(item, h5py.Group):
                    struct[name] = explore(item)
                elif isinstance(item, h5py.Dataset):
                    if item.dtype.names:
                        struct[name] = {
                            field: item.dtype[field].str
                            for field in item.dtype.names
                        }
                    else:
                        struct[name] = tuple(item.shape)
            return struct

        with h5py.File(self.path, "r") as f:
            return explore(f)

    def get_dataset_shape(self, dataset_path: str) -> tuple[int, ...]:
        with h5py.File(self.path, "r") as f:
            obj = self._resolve_path(f, dataset_path)
            return obj.shape

    def load_dataset(self, dataset_path: str) -> np.ndarray:
        with h5py.File(self.path, "r") as f:
            obj = self._resolve_path(f, dataset_path)
            return np.asarray(obj)

    def get_flat_dataset_paths(self) -> list[str]:
        paths = []

        def walk(obj, prefix=""):
            for name, item in obj.items():
                path = f"{prefix}/{name}" if prefix else f"/{name}"
                if isinstance(item, h5py.Group):
                    walk(item, path)
                elif isinstance(item, h5py.Dataset):
                    if item.dtype.names:
                        for field in item.dtype.names:
                            paths.append(f"{path}/{field}")
                    else:
                        paths.append(path)

        with h5py.File(self.path, "r") as f:
            walk(f)

        return paths

    @staticmethod
    def _resolve_path(file, dataset_path: str):
        obj = file
        for part in dataset_path.strip("/").split("/"):
            obj = obj[part]
        return obj



from textual.app import ComposeResult
from textual.widgets import Button, Static, Select
from textual.containers import Vertical

class SettingsTab(Vertical):
    """Widget representing the Settings tab."""

    THEMES_MAP = {
        "Textual Dark": "textual-dark",
        "Textual Light": "textual-light",
        "Nord": "nord",
        "Gruvbox": "gruvbox",
        "Catppuccin Mocha": "catppuccin-mocha",
        "Dracula": "dracula",
        "Tokyo Night": "tokyo-night",
        "Monokai": "monokai",
        "Flexoki": "flexoki",
        "Catppuccin Latte": "catppuccin-latte",
        "Solarized Light": "solarized-light",
    }

    def compose(self) -> ComposeResult:
        yield Static("HEPBrowse Settings", id="settings-title")
        yield Select.from_values(list(self.THEMES_MAP.keys()), prompt="Select Theme", id="theme-select", value="Dracula", allow_blank=False)
        with Vertical(id="save-container"):
            yield Button("Save", id="save-settings", variant="primary")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "save-settings":
            self.notify("Saved settings")

    def on_select_changed(self, event: Select.Changed) -> None:
        """Switch theme using the map."""
        if event.select.id == "theme-select":
            display_name = event.value
            theme_id = self.THEMES_MAP.get(display_name)
            if theme_id:
                self.app.theme = theme_id
                self.notify(f"Switched to {display_name} theme", severity="information")
                self.app.logger.write(f"Set theme to {display_name}")



# src/hepbrowse/helpers/tree_builder.py
from textual.widgets import Tree
from hepbrowse.helpers import loader_factory

class TreeBuilder:
    """Helper class that takes loader output and populates a Textual Tree."""

    @staticmethod
    def populate_from_structure(tree: Tree, structure: dict):
        """Recursively populate a Textual Tree from a nested structure dict."""
        def add_nodes(parent, substruct):
            for key, value in substruct.items():
                if isinstance(value, dict):
                    child = parent.add(key)
                    add_nodes(child, value)
                elif isinstance(value, (list, tuple)):
                    list_node = parent.add(key)
                    for i, item in enumerate(value):
                        list_node.add_leaf(f"[{i}]  {item}")
                else:
                    parent.add_leaf(f"{key}  {value}")
        add_nodes(tree.root, structure)
        tree.root.expand()

    @staticmethod
    def populate_from_file(tree: Tree, file_path: str):
        """Open a data file, build its structure, and fill the tree."""
        loader = loader_factory.get_loader(file_path)
        structure = loader.get_structure()
        TreeBuilder.populate_from_structure(tree, structure)



# src/hepbrowse/data/data_manager.py
from typing import Optional
from hepbrowse.data.loader_base import DataLoader
import numpy as np

class DataManager:
    _instance: Optional["DataManager"] = None  # singleton reference

    def __init__(self, loader: DataLoader):
        self.loader = loader
        self._cache: dict[str, np.ndarray] = {}
        DataManager._instance = self  # set global reference

    @classmethod
    def get(cls) -> "DataManager":
        """Return the global DataManager instance."""
        if cls._instance is None:
            raise RuntimeError("DataManager has not been initialized yet!")
        return cls._instance

    # your previous methods...
    def get_data(self, dataset_path: str) -> np.ndarray:
        if dataset_path in self._cache:
            return self._cache[dataset_path]
        data = self.loader.load_dataset(dataset_path)
        self._cache[dataset_path] = data
        return data

    def get_shape(self, dataset_path: str) -> tuple[int, ...]:
        return self.loader.get_dataset_shape(dataset_path)

    def list_datasets(self) -> list[str]:
        return self.loader.get_flat_dataset_paths()

    def clear_cache(self) -> None:
        self._cache.clear()



from pathlib import Path
from typing import Any
import numpy as np
import uproot
import awkward as ak

from hepbrowse.data.loader_base import DataLoader


class RootLoader(DataLoader):
    """ROOT file loader using uproot (thread-safe, stateless)."""

    def get_structure(self) -> dict:
        """
        Return nested dict of ROOT content:
          - Nested TDirectory structures
          - TTree / TChain names → dict of branch name → interpretation
        """
        import uproot

        structure: dict = {}

        with uproot.open(self.path) as file:
            # uproot <=4 compatibility: no `encoded=` argument
            class_map = file.classnames(recursive=True)

            for fullpath, classname in class_map.items():
                path = fullpath.strip("/")

                if classname == "TDirectory":
                    parts = path.split("/")
                    d = structure
                    for part in parts:
                        d = d.setdefault(part, {})

                elif classname in ("TTree", "TChain"):
                    parts = path.split("/")
                    tree_name = parts[-1]

                    parent = structure
                    for part in parts[:-1]:
                        parent = parent.setdefault(part, {})

                    tree = file[path]

                    branches = {
                        name: str(branch.interpretation)
                        for name, branch in tree.items()
                    }

                    parent[tree_name] = branches

                # TODO: other ROOT object types

        return structure





    # ------------------------------
    # Dataset shape
    # ------------------------------
    def get_dataset_shape(self, dataset_path: str) -> tuple[int, ...]:
        """
        Return shape of a dataset without loading full data.
        Dataset path must be /dir/tree/branch or /tree/branch.
        """
        tree_path, branch = self._split_path(dataset_path)

        with uproot.open(self.path) as file:
            tree = file[tree_path]
            branch_obj = tree[branch]
            return (branch_obj.num_entries,)

    # ------------------------------
    # Dataset loading
    # ------------------------------
    def load_dataset(self, dataset_path: str, *, flatten: bool = True) -> np.ndarray:
        tree_path, branch = self._split_path(dataset_path)
    
        with uproot.open(self.path) as file:
            tree = file[tree_path]
            arr = tree[branch].array(library="ak")
    
            if flatten and "var *" in str(ak.type(arr)):
                return ak.to_numpy(ak.flatten(arr))
            else:
                return ak.to_numpy(arr)


    # ------------------------------
    # Flat dataset listing
    # ------------------------------
    def get_flat_dataset_paths(self) -> list[str]:
        """
        Return a flat list of dataset paths:
          /tree/branch
          /dir/tree/branch
        """
        import uproot

        paths: list[str] = []

        with uproot.open(self.path) as file:
            class_map = file.classnames(recursive=True)

            for fullpath, classname in class_map.items():
                path = fullpath.strip("/")

                if classname in ("TTree", "TChain"):
                    tree = file[path]
                    for branch_name in tree.keys():
                        paths.append(f"/{path}/{branch_name}")

        return paths


    # ------------------------------
    # Helpers
    # ------------------------------
    @staticmethod
    def _split_path(dataset_path: str) -> tuple[str, str]:
        """
        Split /dir/tree/branch → ("dir/tree", "branch")
        """
        parts = dataset_path.strip("/").split("/")
        if len(parts) < 2:
            raise ValueError(f"Invalid ROOT dataset path: {dataset_path}")

        tree_path = "/".join(parts[:-1])
        branch = parts[-1]
        return tree_path, branch



# hepbrowse/data/registry.py
from pathlib import Path
from typing import Type
from hepbrowse.data.loader_base import DataLoader
from hepbrowse.data.h5_loader import H5Loader
from hepbrowse.data.root_loader import RootLoader

_LOADER_REGISTRY: dict[str, Type[DataLoader]] = {
    ".h5": H5Loader,
    ".hdf5": H5Loader,
    ".root": RootLoader
}

@staticmethod
def get_loader(path: str | Path) -> DataLoader:
    path = Path(path)
    suffix = path.suffix.lower()

    if suffix not in _LOADER_REGISTRY:
        raise ValueError(f"No loader registered for file type: {suffix}")

    return _LOADER_REGISTRY[suffix](path)



import importlib
import pkgutil
import sys
import traceback
from typing import List, Type

from hepbrowse.plugins.api import HEPBrowsePlugin


def _iter_plugin_modules(package_name: str):
    """
    Yield imported Python modules under a given package.
    Supports both module-based and package-based plugins.
    """
    try:
        package = importlib.import_module(package_name)
    except Exception:
        return

    package_path = getattr(package, "__path__", None)
    if not package_path:
        return

    for modinfo in pkgutil.iter_modules(package_path, package_name + "."):
        try:
            module = importlib.import_module(modinfo.name)
            yield module
        except Exception:
            print(
                f"[hepbrowse] Failed importing plugin module {modinfo.name}",
                file=sys.stderr,
            )
            traceback.print_exc()


def _find_plugin_classes(module) -> List[Type[HEPBrowsePlugin]]:
    """
    Return all concrete subclasses of HEPBrowsePlugin in a module.
    """
    plugins: List[Type[HEPBrowsePlugin]] = []

    for obj in module.__dict__.values():
        try:
            if (
                isinstance(obj, type)
                and issubclass(obj, HEPBrowsePlugin)
                and obj is not HEPBrowsePlugin
            ):
                plugins.append(obj)
        except Exception:
            continue

    return plugins


def discover_builtin_plugins(
    package: str = "hepbrowse.plugins",
) -> List[HEPBrowsePlugin]:
    """
    Discover and instantiate builtin plugins.
    """
    instances: List[HEPBrowsePlugin] = []

    for module in _iter_plugin_modules(package):
        for plugin_cls in _find_plugin_classes(module):
            try:
                instances.append(plugin_cls())
            except Exception:
                print(
                    f"[hepbrowse] Failed instantiating plugin {plugin_cls.__name__}",
                    file=sys.stderr,
                )
                traceback.print_exc()

    return instances


def register_plugins_with_app(app, package: str = "hepbrowse.plugins"):
    """
    Discover plugins and register them with the app.
    """
    plugins = discover_builtin_plugins(package)

    for plugin in plugins:
        try:
            plugin.register(app)

            if hasattr(plugin, "register_tabs"):
                plugin.register_tabs(app)

            if hasattr(plugin, "register_settings") and hasattr(app, "settings_manager"):
                plugin.register_settings(app.settings_manager)

        except Exception:
            print(
                f"[hepbrowse] Error registering plugin {plugin}",
                file=sys.stderr,
            )
            traceback.print_exc()

    return plugins



from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Label, Input, Select, Button, Static, Markdown, RadioButton
from textual_plotext import PlotextPlot
from textual.worker import Worker
from hepbrowse.plotting.histogram_engine import HistogramEngine

MD_INSTRUCTIONS = """
# Plotting Instructions
Panels on the left can be used to control the plot produced:
- Dataset selection (type to search)
- Number of bins
- Plot limits
- Normalization

Leaving blank fields will result in HEPBrowse automatically defining defaults.

Pressing the **Genenerate Plot** button will load and process the necessary data

Pressing the **Reset Plot** button will erase the displayed plot and show this instructions page again.
"""

class PlotTab(Static):
    """PlotTab Widget"""

    def __init__(self, loader, theme, **kwargs):
        super().__init__(**kwargs)
        self.loader = loader
        self.datasets = self.loader.get_flat_dataset_paths()
        self.theme = theme
        self.display_info = True
        self.plot_area: PlotextPlot | None = None
        self.current_worker: Worker | None = None

    def compose(self) -> ComposeResult:
        with Horizontal():
            with Vertical(id="plot-controls"):

                yield Label("Dataset / Variable")
                yield Select.from_values(
                    self.datasets,
                    prompt="Select Dataset",
                    id="dataset-select",
                    allow_blank=True
                )

                yield Label("Bins")
                self.bins_input = Input(placeholder="50", id="bins-input", type="integer")
                yield self.bins_input

                yield Label("X Min")
                self.xmin_input = Input(placeholder="auto", id="xmin-input", type="number")
                yield self.xmin_input

                yield Label("X Max")
                self.xmax_input = Input(placeholder="auto", id="xmax-input", type="number")
                yield self.xmax_input

                with Horizontal(id="plot-checkboxes"):
                    yield RadioButton("Normalize", id="plot-normalize")

                yield Button("Generate Plot", id="plot-button", variant="success")
                yield Button("Reset Plot", id="reset-button", variant="error")

            with Vertical(id="plot-display"):
                yield Markdown(MD_INSTRUCTIONS, id="plot-md-info")

    async def set_loading(self, loading: bool = True) -> None:
        self.plot_area.loading = loading

    async def set_plot(self, edges, counts, dataset_path) -> None:
        if hasattr(self.plot_area, 'plt'):
            plt = self.plot_area.plt
            plt.bar(edges[:], counts, width=(edges[1] - edges[0]))
            plt.title(f"Histogram for {dataset_path}")
            plt.xlabel("Value")
            plt.ylabel("Count")
            self.plot_area.loading = False

    async def _generate_plot(self):
        if self.plot_area:
            selector = self.query_one("#dataset-select")
            norm_check = self.query_one("#plot-normalize")
            dataset_path = selector.value
            norm = norm_check.value

            try:
                bins = int(self.bins_input.value or 50)
            except ValueError:
                bins = 50

            limits_set = self.xmin_input.value and self.xmax_input.value

            try:
                if limits_set:
                    xmin = float(self.xmin_input.value)
                    xmax = float(self.xmax_input.value)
                    counts, bin_edges = HistogramEngine.make_histogram(dataset_path, bins=bins, range=(xmin, xmax), norm=norm)
                else:
                    counts, bin_edges = HistogramEngine.make_histogram(dataset_path, bins=bins, norm=norm)
            except Exception as e:
                self.app.call_from_thread(
                    self.app.logger.write,
                    f"Histogram generation failed: {e}"
                )
                self.app.call_from_thread(
                    self.notify,
                    f"Error generating histogram: {e}", severity="error"
                )
                self.app.call_from_thread(
                    self.set_loading,
                    False,
                )
                return

            self.app.call_from_thread(
                self.set_plot,
                bin_edges,
                counts,
                dataset_path
            )

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "plot-button":

            selector = self.query_one("#dataset-select")

            if selector.value == Select.BLANK:
                self.notify("No variable was selected for plotting", severity="warning")
                return

            display = self.query_one("#plot-display")

            for child in list(display.children):
                child.remove()

            self.plot_area = PlotextPlot()
            self.plot_area.loading = True
            display.mount(self.plot_area)

            self.current_worker = self.run_worker(self._generate_plot(), group="plotting", exclusive=True, thread=True)
        elif event.button.id == "reset-button":
            display = self.query_one("#plot-display")

            for child in list(display.children):
                child.remove()
            self.plot_area = Markdown(MD_INSTRUCTIONS)
            self.plot_area.loading = False
            display.mount(self.plot_area)


import numpy as np
import plotext as plt
from hepbrowse.data.data_manager import DataManager


class HistogramEngine:
    """Lightweight, NumPy-based histogram engine using lazy data access."""

    @staticmethod
    def make_histogram(dataset_path: str, bins: int = 50, range: tuple[float, float] | None = None, norm: bool = False):
        """
        Create a histogram for the given dataset path.

        This method reopens the file lazily from disk (to ensure thread safety),
        fetches the dataset, computes the histogram with NumPy, and returns both
        bin centers and counts (ready for plotting).

        Parameters
        ----------
        dataset_path : str
            Path to the dataset (as returned by DataManager.list_datasets()).
        bins : int, optional
            Number of histogram bins.
        range : tuple[float, float], optional
            Value range for the histogram; if None, use data min/max.

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            Bin centers and counts arrays.
        """
        # Reopen the file from disk in this thread to avoid threading issues
        dm = DataManager.get()
        loader_cls = type(dm.loader)
        loader_path = dm.loader.path

        # Recreate loader fresh for this thread
        loader = loader_cls(loader_path)

        # Load dataset lazily from the file
        data = loader.load_dataset(dataset_path)

        # Ensure numeric data
        if not np.issubdtype(data.dtype, np.number):
            raise TypeError(f"Dataset '{dataset_path}' is not numeric ({data.dtype}) and cannot be histogrammed.")
        
        # Ensure safe data
        data = data[~np.isnan(data)]
        data = data[np.isfinite(data)]

        # Compute histogram with NumPy
        counts, edges = np.histogram(data, bins=bins, range=range)
        centers = (edges[:-1] + edges[1:]) / 2

        return counts/sum(counts) if norm else counts, centers #centers, counts






# src/hepbrowse/plugins/plot/plugin.py

from hepbrowse.plugins.api import HEPBrowsePlugin
from hepbrowse.widgets.plot_tab import PlotTab

class PlotPlugin(HEPBrowsePlugin):
    name = "Plot"
    version = "0.1"

    def register_tabs(self, app):
        # Use a factory so widget is instantiated during compose, not now.
        app.add_tab("Plot", "plotting", lambda: PlotTab(loader=app.loader, theme=getattr(app, "theme", None)))



from pathlib import Path
from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, Tree, TabbedContent, TabPane, Markdown, RichLog
from hepbrowse.data.data_manager import DataManager
from hepbrowse.settings import SettingsTab
from hepbrowse.helpers.tree_builder import TreeBuilder
from hepbrowse.widgets.plot_tab import PlotTab
from hepbrowse.helpers import loader_factory
from hepbrowse.plugins.loader import register_plugins_with_app

import click

WELCOME_FILE = Path(__file__).parent / "docs" / "welcome.md"

class HEPBrowseApp(App):
    """HEPBrowse main Textual application."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)   

        self.tab_definitions = []   

        self.plugins = []   

        self.services = {}  

        self.plugins = register_plugins_with_app(self)  

    CSS_PATH = [
        Path(__file__).parent / "css" / "hepbrowse.tcss",
        Path(__file__).parent / "css" / "plot_tab.tcss",
        Path(__file__).parent / "css" / "settings_screen.tcss",
        ]
    TITLE = "HEPBrowse"

    BINDINGS = [
        ("ctrl+right", "cycle_tab(1)", "Next Tab"),
        ("ctrl+left", "cycle_tab(-1)", "Previous Tab"),
        ("ctrl+s", "toggle_sidebar", "Toggle Sidebar")
        ]
    
    def set_file(self, path: Path):
        self.file = path
        self.loader = loader_factory.get_loader(path)
        DataManager(self.loader)

    def add_tab(self, label: str, tab_id: str, factory):
        existing = [tid for (_, tid, _) in self.tab_definitions]
        if tab_id in existing:
            return
        self.tab_definitions.append((label, tab_id, factory))

    def _make_welcome_tab(self):
        if WELCOME_FILE.exists():
            return Markdown(WELCOME_FILE.read_text(encoding="utf-8"), id="welcome-text")
        else:
            return Markdown("# Welcome\n\nFile not found.", id="welcome-text")

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield from self.make_sidebar()
        with TabbedContent(initial="welcome"):
            with TabPane("Welcome", id="welcome"):
                if WELCOME_FILE.exists():
                    with WELCOME_FILE.open("r", encoding="utf-8") as f:
                        md_content = f.read()
                else:
                    md_content = "# Welcome\n\nFile not found."
                yield Markdown(md_content, id="welcome-text")
            with TabPane("Plot", id="plotting"):
                yield PlotTab(loader=self.loader, theme=self.theme)
            with TabPane("Settings", id="settings"):
                yield SettingsTab()
            with TabPane("Log", id="log"):
                self.logger = RichLog(highlight=True, markup=True)
                yield self.logger
        yield Footer()
        
    def make_sidebar(self):
        """Generator version of sidebar creation."""
        self.sidebar = Tree(self.file.name, id="sidebar")

        TreeBuilder.populate_from_file(
            self.sidebar,
            str(self.file.resolve())
        )

        yield self.sidebar

    def action_toggle_sidebar(self):
        self.sidebar.toggle_class("--visible")

    def action_cycle_tab(self, direction: int):
        """Move to the next/previous tab with wrapping."""
        tabbed_content = self.query_one(TabbedContent)
        tabs = [tab.id for tab in self.query("TabPane")]
        current_index = tabs.index(tabbed_content.active)
        new_index = (current_index + direction) % len(tabs)
        tabbed_content.active = tabs[new_index]
    
    def on_mount(self) -> None:
        self.logger.write("App starting...")
        self.logger.write("Loading theme...")
        self.theme = "dracula"
        self.logger.write("Loaded file: " + str(self.file.resolve()))
        self.logger.write("App started successfully")

@click.command()
@click.argument('filename', nargs=1)
def HEPBrowse(filename: str):
    """
        Open FILENAME in HEPBrowse 
    """
    app = HEPBrowseApp()
    app.set_file(Path(filename))
    app.run()

if __name__ == "__main__":
    HEPBrowse()